@@ -1,6 +1,7 @@
 from concurrent.futures import ThreadPoolExecutor, as_completed
 from typing import Union, List
-import requests
+from time import time
+import requests, random, webbrowser
 
 from RbxAPI import conversion
 from RbxAPI import utils
@@ -24,8 +25,8 @@ class User(api.BaseAuth):
         Creates an object that provides Roblox user information and endpoint interactions.
 
         :param userid: The id of the user to create an object of.
-        :param cookie: Optional: The user's cookie to use for authentication. This will be required for certain,
-        but not all interactions.
+        :param cookie: Optional: The user's cookie to use for authentication. This will be required for certain, but not all interactions.
+
         :key cookies: Optional: List of cookies to use with a proxy.
         :key proxies: Optional: List of proxies to use with a single cookie or several cookies.
         """
@@ -150,13 +151,13 @@ class User(api.BaseAuth):
         :return: int
         """
         if not self.__rap:
-            data = requests.get(f'{api.inventory}/{self.id}/assets/collectibles?sortOrder=Asc&limit=100').json()
+            data = requests.get(f'{api.inventory}/v1/users/{self.id}/assets/collectibles?sortOrder=Asc&limit=100').json()
             if data.get('errors', ''):
                 utils.handle_code(data['errors'][0]['code'])
             else:
                 results = [data['data']]
                 while data['nextPageCursor']:
-                    data = requests.get(f'{api.inventory}/{self.id}/assets/collectibles?sortOrder=Asc&limit=100&cursor={data["nextPageCursor"]}').json()
+                    data = requests.get(f'{api.inventory}/v1/users/{self.id}/assets/collectibles?sortOrder=Asc&limit=100&cursor={data["nextPageCursor"]}').json()
                     results.append(data['data'])
                 self.__rap = utils.reduce(utils.add, [utils.map_reduce_rap(page) for page in results])
         return self.__rap
@@ -220,6 +221,18 @@ class User(api.BaseAuth):
         else:
             raise UserWarning('Authentication required for this endpoint, session must be set')
 
+    @classmethod
+    def has_asset(cls, userid: int, assetid: int) -> bool:
+        """
+        Class method that returns a boolean based on if the given user owns the given asset.
+
+        :param userid: The id of the user to check ownership for.
+        :param assetid: The id of the asset to check ownership of.
+        :return: bool
+        """
+        data: bool = requests.get(f'{api.base}/ownership/hasasset?userId={userid}&assetId={assetid}').json()
+        return data
+
     @classmethod
     def by_username(cls, username: str) -> 'User':
         """
@@ -318,6 +331,23 @@ class Group(api.BaseAuth):
     def __exit__(self, exc_type, exc_val, exc_tb):
         del self
 
+    def payout(self, type: str, amount: int, recipients: Union[List[User], List[int]]) -> str:
+        """
+        Pays out group funds to the given users.
+
+        :param type: Either a type of `FixedAmount` or `Percentage`. Must enter the type as shown here or it will fail.
+        :param amount: The fixed amount of robux or percentage of robux to pay out to the given users.
+        :return: 'Success'
+        """
+        if isinstance(recipients[0], User):
+            _recipients = [{"recipientId": user.id, "recipientType": "User", "amount": amount} for user in recipients]
+        else:
+            _recipients = [{"recipientId": _id, "recipientType": "User", "amount": amount} for _id in recipients]
+        data = self.session.post(f'{api.groups}/groups/{self.id}/payouts', json={"PayoutType": type, "Recipients": _recipients}).json()
+        if data.get('errors', ''):
+            utils.handle_code(data['errors'][0]['code'])
+        return 'Success'
+
     def update_user_role(self, user: Union[User, int], role: Union[Role, int]) -> str:
         """
         Updates the given user's group role to the given role.
@@ -454,7 +484,7 @@ class Game(api.BaseAuth):
         Creates an object that provides Roblox game information and endpoint interactions.
 
         :param gameid: The id of the game to create an object of.
-        :param cookie: Optional: The user's cookie to use for authentication. This will be required for certain,
+        :param cookie: The user's cookie to use for authentication. This will be required for certain,
         but not all interactions.
         :key cookies: Optional: List of cookies to use with a proxy.
         :key proxies: Optional: List of proxies to use with a single cookie or several cookies.
@@ -474,6 +504,7 @@ class Game(api.BaseAuth):
             data = {k.lower(): v for k, v in resp.items()}
             self.__dict__.update(data)
             self.creator = User(self.creator.get('id')) if self.creator.get('type') == 'User' else Group(self.creator.get('id'))
+            self.__session_ticket = self.session.post('https://auth.roblox.com/v1/authentication-ticket').headers['rbx-authentication-ticket']
             self.__favorites = None
             self.__servers = None
             self.__votes = None
@@ -508,7 +539,10 @@ class Game(api.BaseAuth):
         """
         if not self.__servers:
             data = requests.get(f'{api.games}/games/{self.rootplaceid}/servers/Public?limit=100').json()
-            self.__servers = [Server(server) for server in data['data']]
+            self.__servers = [*data['data']]
+            while data['nextPageCursor']:
+                data = requests.get(f'{api.games}/games/{self.rootplaceid}/servers/Public?limit=100&cursor={data["nextPageCursor"]}').json()
+                self.__servers.append(*data['data'])
         return self.__servers
 
     @property
@@ -522,3 +556,102 @@ class Game(api.BaseAuth):
             data = requests.get(f'{api.games}/games/votes?universeIds={self.id}').json()['data'][0]
             self.__votes = conversion.GameVotes(data['upVotes'], data['downVotes'])
         return self.__votes
+
+    def join(self, lowest_best: bool = False):
+        gameid = None
+        if lowest_best:
+            best = next(filter(lambda s: s.playing == min(self.__servers, key=lambda s: s.playing).playing, self.__servers)).id
+            gameid = f'gameId%3D{best}%26'
+        BrowserID = random.randint(10000000000, 99999999999)
+        webbrowser.open(f"roblox-player:1+launchmode:play+gameinfo:{self.__session_ticket}+launchtime:{int(time()*1000)}+placelauncherurl:https%3A%2F%2Fassetgame.roblox.com%2Fgame%2FPlaceLauncher.ashx%3Frequest%3DRequestGame%26browserTrackerId%3D{BrowserID}%26placeId%3D{self.placeid}%26{gameid}isPlayTogetherGame%3Dfalse+browsertrackerid:{BrowserID}+robloxLocale:en_us+gameLocale:en_us")
+
+
+class Resell:
+    """Class that handles information provided by the economy asset API response."""
+    def __init__(self, data):
+        self.__dict__.update(data)
+        self.seller = User(self.seller['id'])
+
+    def __repr__(self):
+        return f'Resell(seller={self.seller}, price={self.price}, serialNumber={self.serialNumber})'
+
+
+class Asset(api.BaseAuth):
+    """Class that handles interactions with Roblox assets."""
+    def __init__(self, assetid: int, cookie: str = None, **kwargs):
+        """
+        Creates an object that provides Roblox asset information and endpoint interactions.
+
+        :param gameid: The id of the asset to create an object of.
+        :param cookie: Optional: The user's cookie to use for authentication. This will be required for certain,
+        but not all interactions.
+        :key cookies: Optional: List of cookies to use with a proxy.
+        :key proxies: Optional: List of proxies to use with a single cookie or several cookies.
+        """
+        super().__init__(cookie, data=kwargs)
+        if kwargs.get('proxies', ''):
+            requests = self.session
+        else:
+            import requests
+        data = requests.get(f'{api.economy}/{assetid}/resale-data').json()
+        if data.get('errors', ''):
+            utils.handle_code(data['errors'][0]['code'])
+        else:
+            self.__dict__.update(data)
+            self.__sellers = None
+            self.__owners = None
+            self.id = assetid
+
+    def __repr__(self):
+        return f'Asset(id={self.id}, assetStock={self.assetStock}, sales={self.sales}, ' \
+               f'recentAveragePrice={self.recentAveragePrice})'
+
+    def __enter__(self):
+        return self
+
+    def __exit__(self, exc_type, exc_val, exc_tb):
+        del self
+
+    @property
+    def sellers(self) -> List[Resell]:
+        """
+        Contains a list of Resell objects referencing the resellers of the current asset.
+
+        :return: List[Resell]
+        """
+        if not self.__sellers:
+            data = self.session.get(f'{api.economy}/{self.id}/resellers?limit=100').json()
+            if data.get('errors', ''):
+                utils.handle_code(data['errors'][0]['code'])
+            elif not data['data']:
+                self.__sellers = []
+            results = [*data['data']]
+            while data['nextPageCursor']:
+                data = requests.get(f'{api.economy}/{self.id}/resellers?limit=100&cursor={data["nextPageCursor"]}').json()
+                results.append(*data['data'])
+            with ThreadPoolExecutor() as exe:
+                tasks = [exe.submit(Resell, data) for data in results]
+                self.__sellers = [t.result() for t in as_completed(tasks)]
+        return self.__sellers
+
+    @property
+    def owners(self) -> List[User]:
+        """
+        Contains a list of User objects referencing the owners of the current asset.
+
+        :return: List[User]
+        """
+        if not self.__owners:
+            data = self.session.get(f'{api.inventory}/v2/assets/{self.id}/owners?limit=100').json()
+            if data.get('errors', ''):
+                utils.handle_code(data['errors'][0]['code'])
+            elif not data['data']:
+                self.__owners = 'No owners found for this asset'
+            results = [data['data']]
+            while data['nextPageCursor']:
+                data = self.session.get(f'{api.inventory}/v2/assets/{self.id}/owners?limit=100&cursor={data["nextPageCursor"]}').json()
+                results.append(data['data'])
+            with ThreadPoolExecutor() as exe:
+                tasks = [exe.submit(User, data['owner']['id']) for page in results for data in page if data['owner']]
+                self.__owners = [t.result() for t in as_completed(tasks)]
+        return self.__owners
